import numbers
from scipy.stats import mode
import pandas as pd
import numpy as np
import datetime

from mlapp.utils.exceptions.framework_exceptions import UnsupportedFileType


class ClassificationFeatureEngineering(object):
    def drop_features(self, data_df, features_to_drop=None):
        """
        Dropping requested features
        :param data_df: the DataFrame
        :param features_to_drop: list of features names to drop
        :return: data_df after dropping requested featuers
        """
        if not features_to_drop:
            features_to_drop = []

        original_columns = data_df.columns
        filtered_columns_to_drop = filter(lambda x: x in original_columns, features_to_drop)
        return data_df.drop(filtered_columns_to_drop, axis=1)

    def bin_continuous_features(self, data_df, features_to_bin=None):
        """
        Bin continuous features by the configuration in 'features_to_bin'
        :param data_df: the DataFrame
        :param features_to_bin: configuration of bin
        example:

         "features_to_bin":[
          {"name": "feature_name_1", "bins": [5, 15]},
          {"name": "feature_name_2", "bins": [15, 23]}
        ]

        :return: the DataFrame with requested features transformed
        """
        if not features_to_bin:
            features_to_bin = []

        for feature_to_bin in features_to_bin:
            if feature_to_bin['name'] in data_df.columns:
                full_bins = [data_df[feature_to_bin['name']].min() - 1] + feature_to_bin['bins'] + [data_df[feature_to_bin['name']].max() + 1]
                data_df[feature_to_bin['name']] = pd.cut(
                    data_df[feature_to_bin['name']],
                    bins=full_bins,
                    labels=range(len(full_bins) - 1)).astype(float)
        return data_df

    def handle_y_variable(self, data_df, variable_to_predict, options):
        """
        Transform variable to predict by options given in config
        :param data_df: the DataFrame containing all features and variable to predict
        :param variable_to_predict: the variable to predict columns name
        :param options: options containing the configuration of the transformation for the variable to predict
        example:

        "y_variable": {
          "type": "binary",  # binary/multi/continuous - string
          "categories_labels": ["LABEL_1", "LABEL_2"], # category labels - list
          "continuous_to_category_bins": [-0.5, 0.5, 1.5],  # bins values - list
          "label_to_predict": ["LABEL_1"]  # target label to predict - list
        },

        :return: 'data_df' - without the variable to predict, 'final_y' - the variable to predict after transformation
        """
        # y variable configurations
        y_df = data_df[variable_to_predict]
        final_y = pd.DataFrame()
        y_variable_type = options['type']
        target_label = options['label_to_predict']

        # y variable is binary OR one vs all
        if y_variable_type == 'binary' or (y_variable_type == 'multi' and len(target_label) == 1):
            y_dummies = pd.get_dummies(y_df)
            final_y = y_dummies[target_label[0]]

        # y variable is multi class
        elif y_variable_type == 'multi' and len(target_label) < len(y_df.unique()):
            final_y = y_df.apply(lambda x: x if x in target_label else "other")

            # Example for indexing the labels
            # labels_y = final_y.copy()
            # for i in range(len(target_model)):
            #     labels_y = labels_y.apply(lambda x: i + 1 if x == target_model[i] else x)
            # final_y = labels_y.apply(lambda x: 0 if not type(x)==int else x)

        elif y_variable_type == 'continuous':
            bins = options["continuous_to_category_bins"]
            labels = options["categories_labels"]
            final_y = pd.cut(y_df, bins=bins, labels=labels)
        else:
            final_y = y_df

        data_df = data_df.drop(variable_to_predict, axis=1)
        return data_df, final_y

    def transform_and_split_features_to_categorical_and_continuous(self, data, dates_format=None, auto_bin_continuous_features=False, max_categories_num=10):
        """
        Transforming DataFrame features by their value types
        :param data: the DataFrame
        :param dates_format: date formats expected in the DataFrame
        :param auto_bin_continuous_features: whether to bin continuous features automatically
        :param max_categories_num: max unique values in a feature before deciding to auto bin
        :return: the DataFrame with transformed date columns, lists of features by their type, and binned features
        """
        if dates_format is None:
            dates_format = ["%d/%m/%Y", "%Y-%m-%d"]

        data_types = data.dtypes
        today = datetime.datetime.now()
        continuous_columns = []
        continuous_bins = {}
        categorical_columns = []
        binary_columns = []
        for feature, curr_type in data_types.iteritems():
            mysql_type, date_format = self._convert_text_to_date_type(curr_type, feature, data, dates_format)
            if mysql_type == "DATETIME": # converting features from datetime to time_passed_from_date
                data[feature] = data[feature].apply(
                    lambda x: x if self._is_nan(x) else self._elapsed_time_from_date(x, today, date_format))
                if auto_bin_continuous_features:
                    continuous_bins[feature] = np.sort(list(
                        {
                            min(data[feature]) - 1,
                            np.quantile(data[feature].dropna(), 0.2),
                            np.quantile(data[feature].dropna(), 0.4),
                            np.quantile(data[feature].dropna(), 0.6),
                            np.quantile(data[feature].dropna(), 0.8),
                            max(data[feature]) + 1
                        }))
                else:
                    continuous_columns += [feature]
            elif mysql_type == 'NUMERIC':
                unique_values = data[feature].dropna().unique()
                if len(unique_values) == 1:
                    data = data.drop(feature, axis=1)
                elif len(unique_values) == 2:
                    binary_columns += [feature]
                elif (2 < len(unique_values) <= max_categories_num) and auto_bin_continuous_features:
                    categorical_columns += [feature]
                elif auto_bin_continuous_features:
                    continuous_bins[feature] = np.sort(list(
                        {
                            min(data[feature]) - 1,
                            np.quantile(data[feature].dropna(), 0.2),
                            np.quantile(data[feature].dropna(), 0.4),
                            np.quantile(data[feature].dropna(), 0.6),
                            np.quantile(data[feature].dropna(), 0.8),
                            max(data[feature]) + 1
                        }))
                else:
                    continuous_columns += [feature]
            else: # mysql_type == TEXT
                categorical_columns += [feature]

        return data, categorical_columns, continuous_columns, binary_columns, continuous_bins

    def combine_categorical_features(self, data_df, evaluated_df, sep='_|_'):
        """
        Combining categories for each feature
        :param data_df: original DataFrame
        :param evaluated_df: calculated evaluated DataFrame for each category for each feature
        :param sep: separation string
        :return: DataFrame with combined categories
        """
        features_mapping = {}
        results_df = pd.DataFrame()
        groups = pd.DataFrame.groupby(evaluated_df, 'feature_original_name')
        for feature_original_name, group in groups:
            if group.shape[0] > 1:
                # feature_dummies_df = pd.get_dummies(data_df[feature_original_name])

                filtered_feature_dummies_df = data_df[group['feature']]
                combined_feature = filtered_feature_dummies_df.sum(axis=1)

                # preparing feature output name
                categorical_values = group['feature'].apply(lambda x: x.replace(feature_original_name + "_", ""))
                categorical_values = categorical_values.astype(data_df.columns.dtype)
                feature_output_name = feature_original_name + "_"
                for val in categorical_values:
                    feature_output_name += "_" + str(val)

                # adds combined feature to results DataFrame
                results_df[feature_output_name] = combined_feature
            else:
                # save features mappings
                custom_feature_full_name = group['feature'].iloc[0]
                _, new_feature_value = custom_feature_full_name.split(sep)
                features_mapping[feature_original_name] = [{
                    "name": custom_feature_full_name,
                    "categories": [new_feature_value]
                }]

                results_df[group['feature']] = data_df[group['feature']]
        return results_df, features_mapping

    def fillna_features(self, data, features_handling, default_filling=0, missing_values=None):
        """
        Feature handling with filling missing values strategies
        :param data: DataFrame
        :param features_handling: configuration of how to handle each feature
        :return: updated DataFrame with the requested filling
        """
        if missing_values:
            missing_values = {}

        methods = {
            "mean": lambda a: np.mean(a),
            "median": lambda a: np.median(a),
            "mode": lambda a: mode(a).mode[0],
            "none": lambda a: np.nan,
            "nan": lambda a: np.nan
        }

        if not isinstance(data, pd.DataFrame):
            raise UnsupportedFileType("data type should be Dataframe")

        if len(list(missing_values.keys())) > 0:
            data.fillna(missing_values, inplace=True)
        else:
            missing_values = {}
            specific_features = features_handling.keys()
            for feature_key in data.columns:

                # applying fill na on a feature
                if feature_key in specific_features:
                    filling_missing_value = features_handling[feature_key].get("fillna")
                else:
                    filling_missing_value = default_filling

                if filling_missing_value in methods.keys():
                    filling_missing_value = filling_missing_value.lower()
                    val = methods[filling_missing_value](data[feature_key])
                    data[feature_key].fillna(val, inplace=True)
                    missing_values[feature_key] = val
                elif isinstance(filling_missing_value, numbers.Number):
                    data[feature_key].fillna(filling_missing_value, inplace=True)
                    missing_values[feature_key] = filling_missing_value
                else:
                    filling_missing_value = eval(filling_missing_value)
                    if filling_missing_value is None or filling_missing_value == np.nan:
                        data[feature_key] = data[feature_key].fillna(methods["none"], inplace=True)
                        missing_values[feature_key] = np.nan
                    else:
                        val = filling_missing_value(data[feature_key])
                        data[feature_key].fillna(val, inplace=True)
                        missing_values[feature_key] = val
        return data, missing_values


    def transform_features(self, data,features_handling):
        '''
        Feature handling with transformation strategies
        :param data:
        :param features_handling:
        :return: DataFrame - updated DataFrame with the requested transformations
        '''

        if not isinstance(data, pd.DataFrame):
            raise UnsupportedFileType("data type should be Dataframe")

        features = features_handling.keys()
        for feature_key in features:

            # applying transformations
            feature_transformation_methods = features_handling[feature_key].get("transformation", [])
            for feature_transformation_method in feature_transformation_methods:
                data[feature_key] = eval(feature_transformation_method)(data[feature_key])

            # applying dummies
            feature_dummies_flag = features_handling[feature_key].get("dummies", False)
            if feature_dummies_flag:
                dummies_df = pd.get_dummies(data[feature_key], dummy_na=False)
                dummies_df['index'] = data.index.values
                dummies_df = dummies_df.set_index('index')
                data = pd.concat([data, dummies_df], axis=1)
                data = data.drop(feature_key, axis=1)

        return data

    def get_days_from_date(self, data, date_column, extraction_date):
        datenow = datetime.datetime.strptime(extraction_date, '%Y%m%d')
        transformed_data = date_column
        transformed_data = datenow - pd.to_datetime(data[transformed_data], format='%Y%m%d')
        data[date_column] = transformed_data.dt.days
        return data

    def remove_features_by_null_threshold(self, data, percentage=0.3):
        """
        Removing data with amount of 'nulls' more then the 'percentage'
        :param data: the DataFrame
        :param percentage: percentage - default 30%
        :return: pandas DataFrame
        """
        # data = data.loc[:, data.isnull().mean() < percentage]
        n_features = data.shape[1]
        data = data.loc[:, data.isnull().mean() < percentage]
        new_n_features = data.shape[1]
        if n_features == new_n_features:
            print("Features number did not changed, did not found null features more than %0.2f percentage" % percentage)
        else:
            print("%d Features has removed, new data shape is (%d,%d)" % ((n_features - new_n_features), data.shape[0], data.shape[1]))
        return data

    def _is_nan(self, x):
        try:
            return np.isnan(x) or x == ""
        except:
            return False

    def _convert_text_to_date_type(self, data_type, field, csv_file, dates_format=None):
        if dates_format is None:
            dates_format = ["%d/%m/%Y", "%Y-%m-%d"]
        if data_type in ['int64', 'float64', 'float32', 'int32']:
            return "NUMERIC", None
        elif data_type == "object":
            for val in csv_file[field]:
                if val is not None and val is not np.nan:
                    for date_format in dates_format:
                        try:
                            datetime.datetime.strptime(val, date_format)
                            return "DATETIME", date_format
                        except:
                            continue
                    return "TEXT", None
                else:
                    return "TEXT", None
        else:
            return "TEXT", None


    def _elapsed_time_from_date(self, x, today, date_format):
        try:
            return np.round(np.abs((today - datetime.datetime.strptime(str(x), date_format)).days) / 365, 1)
        except:
            return np.nan



basic_regression_config = {
    "pipelines_configs": [
        # train config
        {
            "data_settings": {
                "local_file_path": "data/diabetes.csv",  # data source
                "variable_to_predict": "target",  # column name of variable to predict
            },
            # forecast config
            "model_settings": {
                "local_file_path": "data/diabetes.csv",   # data source
                "variable_to_predict": "target"
            },
            # job settings
            "job_settings": {
                "asset_name": "basic_regression",
                "model_id": "latest",
                "pipeline": "forecast"
            }
        }
    ]
}


from mlapp.handlers.wrappers.wrapper_interface import WrapperInterface


class MessageQueueWrapper(WrapperInterface):
    def init(self):
        """
        Initializes the Wrapper for all handlers of `message_queue` type.
        """
        super(MessageQueueWrapper, self).init('message_queue')

    def send_message(self, queue_name, body):
        """
        Sends message to the queue
        :param queue_name: name of the topic/queue to send the message to
        :param body: message as string or bytes
        """
        if len(self._main_handlers) == 0:
            raise Exception("Called listen to queue without any queue service setup in ML App!")

        for handler_name in self._main_handlers:
            self._handlers[handler_name].send_message(queue_name, body)

    def listen_to_queues(self, queue_names, callback):
        """
        Listen to queues/topics
        :param queue_names: list of queue/topic names to listen to
        :param callback: function to call upon receiving a message
        """
        if len(self._main_handlers) == 0:
            raise Exception("Called listen to queue without any queue service setup in ML App!")

        for handler_name in self._main_handlers:
            self._handlers[handler_name].listen_to_queues(queue_names, callback)


message_queue_instance = MessageQueueWrapper()





import time
from kafka import KafkaConsumer, TopicPartition
from kafka import KafkaProducer
from mlapp.handlers.message_queues.message_queue_interface import MessageQueueInterface


class KafkaHandler(MessageQueueInterface):
    def __init__(self, settings):
        """
        Initializes the KafkaHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(KafkaHandler, self).__init__()

        # init
        self.connection_string = settings.get('hostname') + ':' + str(settings.get('port'))

        self.connection_timeout = settings.get('connection_timeout', 15)

        self.username = settings.get('username')
        self.password = settings.get('password')

        self.security_protocol = "PLAINTEXT"
        self.sasl_mechanism = None

        if self.username != "" and self.password != "":
            self.sasl_mechanism = "PLAIN"
            self.security_protocol = "SASL_PLAINTEXT"

    def send_message(self, queue_name, body):
        """
        Sends message to the queue
        :param queue_name: name of the topic/queue to send the message to
        :param body: message as string or bytes
        """
        producer = KafkaProducer(
            bootstrap_servers=[self.connection_string],
            sasl_plain_username=self.username,
            sasl_plain_password=self.password,
            sasl_mechanism=self.sasl_mechanism,
            security_protocol=self.security_protocol
        )
        b = bytearray()
        b.extend(map(ord, body))
        producer.send(queue_name, value=b, partition=0)
        producer.flush()

    def listen_to_queues(self, queue_names, callback):
        """
        Listen to queues/topics
        :param queue_names: list of queue/topic names to listen to
        :param callback: function to call upon receiving a message
        """
        consumer = KafkaConsumer(
            *queue_names,
            # group_id="my_group",
            bootstrap_servers=[self.connection_string],
            sasl_plain_username=self.username,
            sasl_plain_password=self.password,
            sasl_mechanism=self.sasl_mechanism,
            security_protocol=self.security_protocol
        )

        self.listening = False
        while True:
            try:
                message = consumer.poll()

                for tp, messages in message.items():
                    for message in messages:
                        # message value and key are raw bytes -- decode if necessary!
                        # e.g., for unicode: `message.value.decode('utf-8')`
                        print("%s:%d:%d: key=%s value=%s" % (tp.topic, tp.partition,
                                                             message.offset, message.key,
                                                             message.value))

                        callback(message.value)
                        self.listening = False

                if self.listening == False:
                    self.listening = True
                    print('[*] Waiting for messages in ' + str(queue_names) + '. To exit press CTRL+C')

            except Exception as e:
                print(e)
                time.sleep(1)
                continue




import logging
from mlapp.handlers.file_storages.file_storage_interface import FileStorageInterface
from azure.storage.blob import BlockBlobService


class AzureBlobHandler(FileStorageInterface):
    def __init__(self, settings):
        """
        Initializes the AzureBlobHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(AzureBlobHandler, self).__init__()
        self.connections_parameters = settings
        if self.connections_parameters == {}:
            logging.error('Missing connection parameters!')
        self.block_blob_service = BlockBlobService(account_name=self.connections_parameters['accountName'],
                                                   account_key=self.connections_parameters['accountKey'])

    def download_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        Downloads file from file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        self.block_blob_service.get_blob_to_path(bucket_name, object_name, file_path)

    def stream_file(self, bucket_name, object_name, *args, **kwargs):
        """
        Streams file from file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        return self.block_blob_service.get_blob_to_text(bucket_name, object_name)

    def upload_file(self, bucket_name, object_name, file_path, *args, **kwargs):
        """
        Uploads file to file storage
        :param bucket_name: name of the bucket/container
        :param object_name: name of the object/file
        :param file_path: path to local file
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        # create container (bucket) if it doesn't exist:
        self.block_blob_service.create_container(bucket_name)
        # upload the file:
        self.block_blob_service.create_blob_from_path(bucket_name, object_name, file_path)

    def list_files(self, bucket_name, prefix="", *args, **kwargs):
        """
        Lists files in file storage
        :param bucket_name: name of the bucket/container
        :param prefix: prefix string to search by
        :param args: other arguments containing additional information
        :param kwargs: other keyword arguments containing additional information
        """
        files_names = []

        try:
            blobs = self.block_blob_service.list_blob_names(bucket_name, prefix=prefix)
            for blob in blobs:
                files_names.append(blob)
        except Exception as e:
            logging.error(e)

        return files_names



basic_regression_config = {
    "pipelines_configs": [
        {
            # train config
            "data_settings": {},
            # model settings
            "model_settings": {
                "train_percent": 0.8,
                "variable_to_predict": "target"  # column name of variable to predict
            },
            # task settings
            "job_settings": {
                "asset_name": "basic_regression",
                "reuse_features_id": "latest",
                "pipeline": "reuse_features_and_train"
            }
        }
    ]
}



import os

run_file = \
'''from mlapp import MLApp
from config import settings

mlapp = MLApp(settings)


configs = [
    {
        'config_path': "assets/asset_name/configs/asset_name_train_config.json",  # required
        'asset_name': "asset_name",  # optional
        'config_name': "asset_name_config"  # optional (use with .py files)
    }
]


if __name__ == '__main__':
    for config in configs:
        mlapp.run_flow(config.get('asset_name'), config['config_path'], config.get('config_name'))
'''

app_file = \
'''from mlapp import MLApp
from config import settings

mlapp = MLApp(settings)


if __name__ == '__main__':
    mlapp.run_listener()
'''


utilities_file = \
'''#############################################################
#                                                           #
#  Here you can implement common utilities for your assets  #
#                                                           #
#############################################################
'''


empty_config_file = \
'''settings = {

}
'''


default_config_file = \
'''settings = {
    'env_file_path': "''' + os.path.join('env', '<FILENAME>.env') + '''"
}
'''


docker_compose_file = \
'''version: '3.7'

services:
  mq:
    image: rabbitmq:3-management
    environment:
      RABBITMQ_DEFAULT_USER: "guest"
      RABBITMQ_DEFAULT_PASS: "guest"
      RABBITMQ_DEFAULT_VHOST: "/"
    hostname: "rabbitmq"
    ports:
      - "15673:15672"
      - "5673:5672"
    labels:
      NAME: "rabbitmq"

  db:
    image: postgres:12.1
    restart: always
    environment:
      POSTGRES_USER: 'postgres'
      POSTGRES_PASSWORD: 'mlapp'
      POSTGRES_DB: 'mlapp'
      TZ: 'UTC'
    ports:
      - '5433:5432'
    volumes:
      - postgres-data:/var/lib/postgresql/data
#      - ./init.sql:/docker-entrypoint-initdb.d/init.sql

  fs:
    image: minio/minio:RELEASE.2019-08-21T19-40-07Z
    volumes:
      - minio-data:/data
    ports:
      - "9001:9000"
    environment:
      MINIO_ACCESS_KEY: minio
      MINIO_SECRET_KEY: minio123
    command: server /data
    healthcheck:
      test: ["CMD", "curl", "-f", "http://minio:9000/minio/health/live"]
      interval: 1m30s
      timeout: 20s
      retries: 3
      start_period: 3m

  redis:
    image: bitnami/redis:latest
    volumes:
      - redis-data:/bitnami/redis/data
    environment:
      - ALLOW_EMPTY_PASSWORD=yes
    ports:
      - '6379:6379'

  frontend:
    image: ibmcom/mlapp-cp-ui:latest
    volumes:
      - ./env-config.js:/usr/share/nginx/html/env-config.js
    ports:
      - "8081:8080"

  backend:
    image: ibmcom/mlapp-cp-api:latest
    environment:
      - CORS=http://localhost:8081
      - DB_TYPE=knex
      - DB_ADAPTER=postgres
      - DB_HOST=db
      - DB_USER=postgres
      - DB_PASSWORD=mlapp
      - DB_PORT=5432
      - DB_NAME=mlapp
      - FS_TYPE=minio
      - FS_ENDPOINT=fs
      - FS_ACCESSKEY=minio
      - FS_SECRETKEY=minio123
      - FS_PORT=9000
      - MQ_TYPE=rabbitmq
      - MQ_ENDPOINT=amqp://guest:guest@mq:5672
      - APP_LOGIN_REQUIRED=false
      - APP_IS_HTTPS=false
      - APP_LOGIN_TYPE=basic
      - SESSION_TYPE=redis
      - REDIS_HOST=127.0.0.1
      - REDIS_PORT=6379
      - CLIENT_API_TOKEN=<YOUR_TOKEN>
      - APP_PKEY=<YOUR_PKEY>
    ports:
      - "3001:3000"
    depends_on:
      - "db"
      - "mq"
      - "fs"
      - "redis"

volumes:
  minio-data:
  postgres-data:
  redis-data:
'''

init_sql_file = \
'''CREATE TABLE public.analysis_results (
	model_id uuid NOT NULL,
	asset_name varchar(255) NOT NULL,
	asset_label varchar(255) NULL,
	pipeline varchar(255) NOT NULL,
	properties json NOT NULL,
	metadata json NOT NULL,
	environment varchar(255) NULL,
	created_at timestamptz NULL,
	CONSTRAINT analysis_results_pkey PRIMARY KEY (model_id)
);

CREATE TABLE public.asset_accuracy_monitoring (
	model_id uuid NOT NULL,
	asset_name varchar(255) NOT NULL,
	asset_label_name varchar(255) NOT NULL,
	created_at timestamptz NULL,
	updated_at timestamptz NULL,
	"timestamp" timestamptz NULL DEFAULT CURRENT_TIMESTAMP,
	model_accuracy json NOT NULL
);

CREATE TABLE public.flows (
	flow_id uuid NOT NULL,
	metadata json NOT NULL,
	properties json NOT NULL,
	created_at timestamptz NULL DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE public.target (
	"timestamp" timestamptz NOT NULL,
	model_id uuid NOT NULL,
	forecast_id uuid NOT NULL,
	"index" varchar(255) NOT NULL,
	y_true float4 NULL,
	y_hat float4 NULL,
	"type" int4 NOT NULL,
	CONSTRAINT target_pkey PRIMARY KEY (model_id, forecast_id, index, type)
);

CREATE TABLE public.jobs (
	id uuid NOT NULL,
	"user" varchar(255) NOT NULL,
	"data" json NULL,
	status_code int4 NULL,
	status_msg varchar(1028) NULL,
	created_at timestamptz NULL,
	updated_at varchar(255) NULL,
	CONSTRAINT tasks_pkey PRIMARY KEY (id)
);
'''

vue_env_config_file = \
'''var env_config = (() => {
  return {
    "VUE_APP_BASE_URL": "http://localhost:3001",
    "VUE_APP_FILE_STORAGE_URL": "http://127.0.0.1:9001",
    "VUE_APP_LOGIN_REQUIRED": "false",
    "VUE_APP_LOGIN_TYPE": "basic",
    "VUE_APP_LOGIN_BACKGROUND": "",
    "VUE_APP_LOGO": "https://repository-images.githubusercontent.com/342536969/591cbb00-7c38-11eb-94ce-e2a2f77039e3",
    "VUE_APP_DEPLOYMENT": "default",
    "FILE_STORE_BUCKETS": "imgs:mlapp-imgs,configs:mlapp-configs,logs:mlapp-logs"
  };
})();
'''

env_file = \
'''# RabbitMQ
RABBITMQ_MLAPP_SERVICE_TYPE=rabbitmq
RABBITMQ_MAIN_MQ=true
RABBITMQ_RABBITMQ_HOSTNAME=127.0.0.1
RABBITMQ_RABBITMQ_PORT=5673
RABBITMQ_RABBITMQ_CONNECTION_TIMEOUT=15

# Postgres
POSTGRES_MLAPP_SERVICE_TYPE=postgres
POSTGRES_MAIN_DB=true
POSTGRES_POSTGRES_DATABASE_NAME=mlapp
POSTGRES_POSTGRES_HOSTNAME=127.0.0.1
POSTGRES_POSTGRES_PORT=5433
POSTGRES_POSTGRES_USER_ID=postgres
POSTGRES_POSTGRES_PASSWORD=mlapp

# Minio
MINIO_MLAPP_SERVICE_TYPE=minio
MINIO_MAIN_FS=true
MINIO_MINIO_ENDPOINT=127.0.0.1
MINIO_MINIO_ACCESS_KEY=minio
MINIO_MINIO_SECRET_KEY=minio123
MINIO_MINIO_PORT=9001
'''


azureml_env_file = \
'''# AML Storage handler
AML-STORAGE_MLAPP_SERVICE_TYPE=azureml_model_storage
AML-STORAGE_MAIN_FS=true

# azureml-queue azureml-queue service
AZUREML-QUEUE_MLAPP_SERVICE_TYPE=azureml_queue
AZUREML-QUEUE_MAIN_MQ=true
'''


# This file expects for asset_name argument.
model_manager_file = \
'''from mlapp.managers import ModelManager, pipeline


class {}ModelManager(ModelManager):
    @pipeline
    def train_model(self, data):
        raise NotImplementedError()

    @pipeline
    def forecast(self, data):
        raise NotImplementedError()
    
    @pipeline
    def refit(self, data):
        raise NotImplementedError()
'''

# This file expects for asset_name argument.
data_manager_file = \
'''from mlapp.managers import DataManager, pipeline


class {}DataManager(DataManager):
    # -------------------------------------- train pipeline -------------------------------------------
    @pipeline
    def load_train_data(self, *args):
        raise NotImplementedError("should return data")
    
    @pipeline
    def clean_train_data(self, data):
        return data
    
    @pipeline
    def transform_train_data(self, data):
        return data

    # ------------------------------------ forecast pipeline -------------------------------------------
    @pipeline
    def load_forecast_data(self, *args):
        raise NotImplementedError("should return data")
    
    @pipeline
    def clean_forecast_data(self, data):
        return data

    @pipeline
    def transform_forecast_data(self, data):
        return data
    
    @pipeline
    def load_target_data(self, *args):
        raise NotImplementedError()

'''

# This file expects to replace `<ASSET_NAME>` in asset_name.
train_config_file = \
'''{
  "pipelines_configs": [
    {
      "data_settings": {

      },
      "model_settings": {

      },
      "job_settings": {
        "asset_name": "<ASSET_NAME>",
        "pipeline": "train"
      }
    }
  ]
}
'''

# This file expects to replace `<ASSET_NAME>` in asset_name.
train_config_file_with_flow = \
'''{
  "pipelines_configs": [
    {
      "data_settings": {
        
      },
      "model_settings": {
        
      },
      "job_settings": {
        "asset_name": "<ASSET_NAME>",
        "pipeline": "train"
      }
    }
  ],
  "flow_config": {
    
  },
  "job_settings": {
    "task_type": "flow"
  }
}
'''

# This file expects to replace `<ASSET_NAME>` in asset_name.
forecast_config_file = \
'''{
  "pipelines_configs": [
    {
      "data_settings": {
        
      },
      "model_settings": {
        
      },
      "job_settings": {
        "asset_name": "<ASSET_NAME>",
        "pipeline": "forecast",
        "model_id": "latest"
      }
    }
  ]
}
'''

# This file expects to replace `<ASSET_NAME>` in asset_name.
forecast_config_file_with_flow = \
'''{
  "pipelines_configs": [
    {
      "data_settings": {

      },
      "model_settings": {
      },
      "job_settings": {
        "asset_name": "<ASSET_NAME>",
        "pipeline": "forecast",
        "model_id": "latest"
      }
    }
  ],
  "flow_config": {},
  "job_settings": {
    "task_type": "flow"
  }
}
'''


gitignore_file = \
'''*.idea/
*.pyc
.DS_Store
my-vern
my_venv
*.log
*.pkl
*.csv
output/
temporary_output/
test_output/
metastore_db/
cache/
venv/
dist/
build/
*egg-info
*.env*
.env*
'''

dockerignore_file = \
'''.dockerignore
Dockerfile
db.sqlite3
__pycache__
*.pyc
*.pyo
*.pyd
.Python
env
pip-log.txt
pip-delete-this-directory.txt
.tox
.coverage
.coverage.*
.cache
coverage.xml
*,cover
*.log
.git
*.idea/
.DS_Store
my-vern
my_venv
venv/
dist/
output/
temporary_output/
test_output/
build/
*egg-info
telepresence.log
'''

amlignore_file = '''*.idea/
*.pyc
.DS_Store
my-vern
my_venv
output/
temporary_output/
test_output/
metastore_db/
cache/
venv/
dist/
build/
*egg-info

'''


from mlapp.handlers.message_queues.message_queue_interface import MessageQueueInterface
from azure.servicebus import ServiceBusClient, ReceiveSettleMode
from azure.servicebus import Message
from azure.servicebus.common.errors import ServiceBusError
import time


class AzureServicesBusHandler(MessageQueueInterface):
    def __init__(self, settings):
        """
        Initializes the AzureServicesBusHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(AzureServicesBusHandler, self).__init__()

        # init
        self.client = ServiceBusClient(
            service_namespace=settings.get('hostname'),
            shared_access_key_name=settings.get('shared_access_key_name'),
            shared_access_key_value=settings.get('shared_access_key'),
            debug=False)

        self.connection_timeout = settings.get('connection_timeout', 10)

    def send_message(self, queue_name, body):
        """
        Sends message to the queue
        :param queue_name: name of the topic/queue to send the message to
        :param body: message as string or bytes
        """
        while True:
            try:
                queue_client = self.client.get_queue(queue_name)
                with queue_client.get_sender() as sender:
                    message = Message(body)
                    sender.send(message)
                break
            except Exception as e:
                print(e)

    def listen_to_queues(self, queue_names, callback):
        print('[*] Waiting for messages in ' + str(queue_names[0]) + '. To exit press CTRL+C')
        while True:
            try:
                queue_client = self.client.get_queue(queue_names[0])
                with queue_client.get_receiver(mode=ReceiveSettleMode.PeekLock) as receiver:
                    batch = receiver.fetch_next(max_batch_size=1, timeout=self.connection_timeout)
                    while batch:
                        for message in batch:
                            message.complete()
                            callback(str(message))
                        batch = receiver.fetch_next(max_batch_size=1, timeout=self.connection_timeout)
            except ServiceBusError as e:
                print(str(e))
            finally:
                time.sleep(1)



from mlapp.handlers.wrappers.wrapper_interface import WrapperInterface


class DatabaseWrapper(WrapperInterface):
    def init(self):
        """
        Initializes the Wrapper for all handlers of `database` type.
        """
        super(DatabaseWrapper, self).init('database')

    def execute_query(self, query, params=None):
        """
        Executes Query in the database.
        :param query: str - query to be executed.
        :param params: list - list of parameters to be used if necessary in query
        :return: result of query
        """
        for handler_name in self._main_handlers:
            result = self._handlers[handler_name].execute_query(query, params)
            if hasattr(result, '__len__'):
                return result

    def insert_query(self, query, values):
        """
        Executes an "INSERT" query in the database.
        :param query: str - query to be executed.
        :param values: list - list of values to be used in the query
        """
        for handler_name in self._main_handlers:
            self._handlers[handler_name].insert_query(query, values)

    def insert_df(self, sql_table, df, batch_length=1000):
        """
        Inserts a DataFrame into a table in the database.
        :param sql_table: str - name of the table.
        :param df: DataFrame (Pandas, PySpark or other) - Matrix type DataFrame containing all values to insert.
        :param batch_length: int - length of the how many rows to insert from matrix at a time
        """
        for handler_name in self._main_handlers:
            self._handlers[handler_name].insert_df(sql_table, df, batch_length)

    def get_df(self, query, params=None):
        """
        Executes a query in the database and returns it as a DataFrame.
        :param query: str - query to be executed.
        :param params: list - list of parameters to be used if necessary in query
        :return: result of query as a DataFrame
        """
        for handler_name in self._main_handlers:
            return self._handlers[handler_name].get_df(query, params)

    def update_job_running(self, job_id):
        """
        Updates row in the table of jobs by the job_id to status `Running`
        Functionality of the Control Panel
        :param job_id: str - id of the job
        """
        for handler_name in self._main_handlers:
            self._handlers[handler_name].update_job_running(job_id)

    def update_actuals(self, df):
        """
        Update target table with the y_true
        :param df: the dataframe that represents the real data that was loaded
        :return: none
        """
        for handler_name in self._main_handlers:
            self._handlers[handler_name].update_actuals(df)

    def get_model_predictions(self, model_id, prediction_type=3, from_date=None, to_date=None):
        """
        Update target table with the y_true
        :param df: the dataframe that represents the real data that was loaded
        :return: none
        """
        for handler_name in self._main_handlers:
            return self._handlers[handler_name].get_model_predictions(
                model_id, prediction_type=3, from_date=None, to_date=None
            )


database_instance = DatabaseWrapper()



from abc import ABCMeta, abstractmethod


class MessageQueueInterface:
    __metaclass__ = ABCMeta

    @abstractmethod
    def send_message(self, queue_name, body):
        """
        Sends message to the queue
        :param queue_name: name of the topic/queue to send the message to
        :param body: message as string or bytes
        :return: None
        """
        raise NotImplementedError()

    @abstractmethod
    def listen_to_queues(self, queue_names, callback):
        """
        Listen to queues/topics
        :param queue_names: list of queue/topic names to listen to
        :param callback: function to call upon receiving a message
        :return: None
        """
        raise NotImplementedError()




import imp
import os
from setuptools import setup, find_packages
from os import path

# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import
from io import open

version = imp.load_source('mlapp.version', os.path.join('mlapp', 'version.py')).VERSION
here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

with open("requirements.txt", "r", encoding="utf-8") as f:
    requires = [x.strip() for x in f if x.strip()]

setup(
    name='mlapp',  # Required
    version=version,  # Required
    description='IBM Services Framework for ML Applications Python 3 framework for building robust, '
                'production-ready machine learning applications. '
                'Official ML accelerator within the larger RAD-ML methodology.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ibm/mlapp',
    author='IBM',
    author_email='tomer.galula@ibm.com, tal.waitzenberg@ibm.com, michael.chein@ibm.com, erez.nardia@ibm.com, '
                 'annaelle.cohen@ibm.com, katzn@us.ibm.com',
    keywords=['mlapp', 'ibm', 'machine-learning', 'auto-ml'],
    packages=find_packages(exclude=['tests', 'docs', 'venv']),  # Required
    license="Apache License 2.0",
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: Apache Software License'
    ],
    python_requires='>=3.6',

    install_requires=requires,

    extras_require={  # Optional
        'rabbitmq': ['pika'],
        'minio': ['minio'],
        'mysql': ['PyMySQL'],
        'snowflake': ['snowflake-sqlalchemy'],
        'azure-servicebus': ['azure-servicebus<=0.50.3'],
        'kafka': ['kafka-python'],
        'kafka-kerberos': ['confluent_kafka'],
        'boto3': ['boto3'],
        'ibm-boto3': ['ibm-cos-sdk'],
        'azure-storage-blob': ['azure-storage-blob<=2.1.0'],
        'postgres': ['pg8000<=1.16.5'],
        'livy': ['livy'],
        'mssql': ['pyodbc'],
        'pyspark': ['pyspark'],
        'aml': [
            'azureml-sdk'
        ],
        'cp': [
            'pika',
            'pg8000<=1.16.5',
            'minio'
        ]
    },
    # package_data={
    #     'mlapp': ['mlapp/cli/script.sh'],
    # },
    # data_files=[('my_data', ['data/data_file'])],  # Optional
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'mlapp=mlapp.cli:cli',
        ],
    },

    project_urls={
        'Bug Reports': 'https://github.com/ibm/mlapp/issues',
        'Wiki Page': 'https://mlapp-docs.s3-web.us-south.cloud-object-storage.appdomain.cloud',
        'Crash Course': 'https://mlapp-docs.s3-web.us-south.cloud-object-storage.appdomain.cloud'
                        '/crash-course/introduction',
        'Source': 'https://github.com/ibm/mlapp'
    },
)



import datetime
import numbers
from pyspark.sql import DataFrame
import pyspark.sql.functions as F
import pandas as pd
import numpy as np

from mlapp.utils.features.spark import spark_cut, spark_dummies, spark_select_dummies
from mlapp.utils.exceptions.framework_exceptions import UnsupportedFileType


class SparkClassificationFeatureEngineering(object):
    def drop_features(self, data_df, features_to_drop=[]):
        """
        Dropping requested features
        :param data_df: the DataFrame
        :param features_to_drop: list of features names to drop
        :return: data_df after dropping requested featuers
        """
        original_columns = data_df.columns
        filtered_columns_to_drop = filter(lambda x: x in original_columns, features_to_drop)
        return data_df.drop(filtered_columns_to_drop, axis=1)

    def bin_continuous_features(self, data_df, features_to_bin=[]):
        """
        Bin continuous features by the configuration in 'features_to_bin'
        :param data_df: the spark DataFrame
        :param features_to_bin: configuration of bin
        example:

         "features_to_bin":[
          {"name": "feature_name_1", "bins": [5, 15]},
          {"name": "feature_name_2", "bins": [15, 23]}
        ]
        if bins is of length 1 containing one integer n, the feature will be binned to n equal sections.
        if bins is == 'quantile' the feature will be divided according to the quantiles - 0.2, 0.4, 0.6, 0.8
        :return: the DataFrame with requested features transformed
        """
        for feature_to_bin in features_to_bin:
            if feature_to_bin['name'] in data_df.columns:
                if len(feature_to_bin['bins']) == 1:
                    if feature_to_bin['bins'][0] == 'quantile':
                        # bin to quantiles.
                        quantiles = [0.2, 0.4, 0.6, 0.8]
                        min_val = data_df.agg({feature_to_bin['name']: 'min'}).collect()[0][0] -1
                        max_val = data_df.agg({feature_to_bin['name']: 'max'}).collect()[0][0] +1
                        q = data_df.approxQuantile(feature_to_bin['name'], quantiles, relativeError=0.05) #relativeError of 0 is more accurate but more computationaly expensive.
                        bins = [min_val] + q + [max_val]
                        data_df = spark_cut(data_df, feature_to_bin['name'], bins=bins,
                                                 labels=[f'Q_{str(q)[2]}0%' for q in quantiles]+['Q100%'])

                    if isinstance(feature_to_bin['bins'][0], int):
                    # an integer (n) was provided. bin to n identical sections.
                        if len(data_df.select(feature_to_bin['name']).distinct().collect()) > feature_to_bin['bins'][0]:
                            min_val = data_df.agg({feature_to_bin['name']: 'min'}).collect()[0][0] -1
                            max_val = data_df.agg({feature_to_bin['name']: 'max'}).collect()[0][0] +1
                            bins = np.linspace(min_val, max_val, feature_to_bin['bins'][0])
                            data_df = spark_cut(data_df, feature_to_bin['name'], bins=bins,
                                                     labels=range(feature_to_bin['bins'][0]-1))
                else:
                    # bin according to the provided bins.
                    min_val = data_df.agg({feature_to_bin['name']: "min"}).collect()[0][0]
                    max_val = data_df.agg({feature_to_bin['name']: "max"}).collect()[0][0]
                    full_bins = sorted([(min_val - 1)] + feature_to_bin['bins'] + [(max_val + 1)])
                    data_df = spark_cut(
                        data_df, feature_to_bin['name'],
                        bins=full_bins,
                        labels=range(len(full_bins) - 1))
        return data_df

    def handle_y_variable(self, data_df, variable_to_predict, options):
        """
        Transform variable to predict by options given in config
        :param data_df: the DataFrame containing all features and variable to predict
        :param variable_to_predict: the variable to predict columns name
        :param options: options containing the configuration of the transformation for the variable to predict
        example:

        "y_variable": {
          "type": "binary",  # binary/multi/continuous - string
          "categories_labels": ["LABEL_1", "LABEL_2"], # category labels - list
          "continuous_to_category_bins": [-0.5, 0.5, 1.5],  # bins values - list
          "label_to_predict": ["LABEL_1"]  # target label to predict - list
        },

        :return: 'data_df' - without the variable to predict, 'final_y' - the variable to predict after transformation
        """
        # y variable configurations
        y_variable_type = options['type']
        target_label = options['label_to_predict']

        # y variable is binary OR one vs all
        if y_variable_type == 'binary' or (y_variable_type == 'multi' and len(target_label) == 1):
            data_df = spark_dummies(data_df, variable_to_predict)
            data_df = spark_select_dummies(data_df, variable_to_predict, target_label)
            y_column_name = f'{variable_to_predict}_{target_label[0]}'
            data_df = data_df.withColumn(variable_to_predict, F.col(y_column_name)).drop(y_column_name)

        # y variable is multi class
        elif y_variable_type == 'multi' and len(target_label) < len(data_df.select(variable_to_predict).distinct().collect()[0][0]):
            data_df = data_df.withColumn(
                variable_to_predict,
                F.when(F.col(variable_to_predict).isin(target_label), data_df[variable_to_predict]).otherwise('other')
            )

        # y variable continuous
        elif y_variable_type == 'continuous':
            bins = options["continuous_to_category_bins"]
            labels = options["categories_labels"]

            data_df = spark_cut(data_df, variable_to_predict,  bins=bins, labels=labels)

        return data_df

    def transform_dates_to_elapsed_days(self, data, dates_format):
        """
        Go over all features and convert any datetime features to elapsed days till today's date.
        :param data:  spark DataFrame
        :param dates_format: date formats expected in the DataFrame
        :return: the DataFrame with transformed date columns.
        """
        if dates_format is None:
            dates_format = 'yyyyMMdd'

        data_types = data.dtypes
        today = datetime.datetime.now()
        for feature, dtype in data_types:
            if dtype == 'timestamp':
                # Convert timestamp to date.
                data = data.withColumn(feature, F.to_date(feature, dates_format))
                # Convert to elapsed days:
                data.withColumn(feature, F.datediff(F.to_date(F.lit(today)),
                                              F.to_date(feature, "yyyy/MM/dd")))
        return data

    def combine_categorical_features(self, data_df, evaluated_df, sep = '_|_'):
        """
        Combining categories for each feature
        :param data_df: original DataFrame
        :param evaluated_df: calculated evaluated DataFrame for each category for each feature
        :return: DataFrame with combined categories
        """
        features_mapping = {}
        results_df = pd.DataFrame()
        groups = pd.DataFrame.groupby(evaluated_df, 'feature_original_name')
        for feature_original_name, group in groups:
            if group.shape[0] > 1:
                # feature_dummies_df = pd.get_dummies(data_df[feature_original_name])

                filtered_feature_dummies_df = data_df[group['feature']]
                combined_feature = filtered_feature_dummies_df.sum(axis=1)

                # preparing feature output name
                categorical_values = group['feature'].apply(lambda x: x.replace(feature_original_name + "_", ""))
                categorical_values = categorical_values.astype(data_df.columns.dtype)
                feature_output_name = feature_original_name + "_"
                for val in categorical_values:
                    feature_output_name += "_" + str(val)

                # adds combined feature to results DataFrame
                results_df[feature_output_name] = combined_feature
            else:
                # save features mappings
                custom_feature_full_name = group['feature'].iloc[0]
                _, new_feature_value = custom_feature_full_name.split(sep)
                features_mapping[feature_original_name] = [{
                    "name": custom_feature_full_name,
                    "categories": [new_feature_value]
                }]

                results_df[group['feature']] = data_df[group['feature']]
        return results_df, features_mapping

    def fillna_features(self, data, features_handling, default_filling=0, stored_missing_values={}):
        """
        Feature handling with filling missing values strategies
        :param data: DataFrame
        :param features_handling: configuration of how to handle each feature
        :return: updated DataFrame with the requested filling
        """
        methods = {
            "mean": lambda a: F.mean(a),
            "median": lambda a: data.approxQuantile(feature_key,[0.5],0.25)[0],
            # "mode": lambda a: mode(a).mode[0],
            "none": lambda a: float('nan'),
            "nan": lambda a: float('nan')
        }

        if not isinstance(data, DataFrame):
            raise UnsupportedFileType("Error: data type should be spark Dataframe")

        if len(list(stored_missing_values.keys())) > 0:
            data = data.fillna(stored_missing_values)
            missing_values = stored_missing_values
        else:
            missing_values = {}

            specific_features = features_handling.keys()
            for feature_key in data.columns:

                # applying fillna on a feature
                if feature_key in specific_features:
                    filling_missing_value = features_handling[feature_key].get("fillna")
                else:
                    filling_missing_value = default_filling

                if filling_missing_value in methods.keys():
                    filling_missing_value = filling_missing_value.lower()
                    val = data.select(methods[filling_missing_value](data[feature_key])).collect()[0][0]
                    data = data.fillna({feature_key: val})
                    missing_values[feature_key] = val
                elif isinstance(filling_missing_value, numbers.Number):
                    data = data.fillna({feature_key: filling_missing_value})
                    missing_values[feature_key] = filling_missing_value
                else:
                    filling_missing_value = eval('F.' + filling_missing_value)
                    if filling_missing_value is None or np.isnan(filling_missing_value):
                        data = data.fillna({feature_key: float('nan')})
                        missing_values[feature_key] = float('nan')
                    else:
                        val = data.agg(filling_missing_value(feature_key)).collect()[0][0]
                        data = data[feature_key].fillna(val)
                        missing_values[feature_key] = val
        return data, missing_values

    def transform_features(self, data, features_handling):
        '''
        Feature handling with transformation strategies
        :param data:
        :param features_handling:
        :return: DataFrame - updated DataFrame with the requested transformations
        '''

        if not isinstance(data, DataFrame):
            raise p("Error: data type should be a spark Dataframe")

        features = features_handling.keys()
        for feature_key in features:

            # applying transformations
            feature_transformation_methods = features_handling[feature_key].get("transformation", [])
            for feature_transformation_method in feature_transformation_methods:
                data = data.withColumn(feature_key, eval(feature_transformation_method)(data[feature_key]))

            # applying dummies
            feature_dummies_flag = features_handling[feature_key].get("dummies", False)
            if feature_dummies_flag:
                data = spark_dummies(data, feature_key)

        return data

    def get_days_from_date(self, data, date_column, extraction_date):
        datenow = datetime.datetime.strptime(extraction_date, '%Y%m%d')
        transformed_data = date_column
        transformed_data = datenow - pd.to_datetime(data[transformed_data], format='%Y%m%d')
        data[date_column] = transformed_data.dt.days
        return data

    def remove_features_by_null_threshold(self, data, percentage=0.3):
        """
        Removing data with amount of 'nulls' more then the 'percentage'
        :param data: the DataFrame
        :param percentage: percentage - default 30%
        :return: pandas DataFrame
        """
        null_percentages = data.select(
            [(F.count(F.when(F.isnull(c), c)) / data.count()).alias(c) for c in data.columns]).collect()[0]

        n_features = len(data.columns)
        data = data.select([c for c in data.columns if null_percentages[c] < percentage])
        new_n_features = len(data.columns)
        if n_features == new_n_features:
            print("Features number was not changed, did not found null features more than %0.2f percentage" % percentage)
        else:
            print("%d Features has removed, new data shape is (%d,%d)" % ((n_features - new_n_features), data.shape[0], data.shape[1]))
        return data

    def _is_nan(self, x):
        try:
            return np.isnan(x) or x == ""
        except:
            return False

    def _convert_text_to_date_type(self, data_type, field, csv_file, dates_format=None):
        if dates_format is None:
            dates_format = ["%d/%m/%Y", "%Y-%m-%d"]
        if data_type in ['int64', 'float64', 'float32', 'int32']:
            return "NUMERIC", None
        elif data_type == "object":
            for val in csv_file[field]:
                if val is not None and val is not np.nan:
                    for date_format in dates_format:
                        try:
                            datetime.datetime.strptime(val, date_format)
                            return "DATETIME", date_format
                        except:
                            continue
                    return "TEXT", None
                else:
                    return "TEXT", None
        else:
            return "TEXT", None

    def _elapsed_time_from_date(self, x, today, date_format):
        try:
            return np.round(np.abs((today - datetime.datetime.strptime(str(x), date_format)).days) / 365, 1)
        except:
            return np.nan


    def convert_features_to_dummies(self, data, features):
        for feature in features:
            if feature in list(data.columns):
                data = spark_dummies(data, feature)
        return data

    def interact_features(self, data, interact_list):
        for features_pair in interact_list:
            feature_0 = features_pair[0]
            feature_1 = features_pair[1]

            if (feature_0 not in data.columns) or (feature_1 not in data.columns):
                print('Warning: one the features: ' + feature_0 + ',' + feature_1 + 'do not exists in the Data.')
            else:
                name = feature_0 + '_' + feature_1
                data = data.withColumn(name, (F.col(feature_0)) * (F.col(feature_0)))
        return data




import numpy as np
from pyspark.ml.feature import VectorAssembler
from mlapp.managers import ModelManager, pipeline
from mlapp.utils.automl import AutoMLSpark


class SparkRegressionModelManager(ModelManager):
    @pipeline
    def train_model(self, data):
        # preparation for train
        variable_to_predict = self.model_settings['variable_to_predict']

        # split the data to test and train
        train_percent = self.model_settings.get('train_percent', 0.8)
        train_data, test_data = data.randomSplit([train_percent, (1 - train_percent)])

        # regression
        result = AutoMLSpark('linear').run(train_data, test_data, variable_to_predict, cv=3)

        # store results
        self.save_automl_result(result, obj_type='pyspark')

    @pipeline
    def forecast(self, data_df):
        result = self.get_automl_result()

        # get best model
        model = result.get_best_model()

        # get selected features from feature selection algorithm
        selected_features_names = result.get_selected_features()

        # filter data according to selected features
        filtered_data = data_df.select(selected_features_names)

        # assemble data for train
        assembler = VectorAssembler(inputCols=selected_features_names, outputCol='features')
        data_df = assembler.transform(filtered_data).select('features')

        # prediction
        predictions = model.transform(data_df)
        num_of_predictions = predictions.count()

        # adding prediction to results
        self.add_predictions(np.arange(num_of_predictions),                                 # index array
                             predictions.select('prediction').toPandas()['prediction'],     # y_hat
                             np.array([np.nan] * num_of_predictions),                       # y_true (empty array)
                             'forecast')                                                    # forecast type

    @pipeline
    def refit(self, data):
        raise NotImplementedError()



from pyspark.mllib.evaluation import MulticlassMetrics, BinaryClassificationMetrics
import pyspark.sql.functions as F
from pyspark.ml.evaluation import Evaluator, RegressionEvaluator, MulticlassClassificationEvaluator, \
    BinaryClassificationEvaluator


def classification(train_predictions, test_predictions, train_actuals=None, test_actuals=None,
                   variable_to_predict='target', prediction_col='prediction', *args, **kwargs):
    """
    Return a dictionary of accuracy scores for provided predicted values. The following scores are returned:
    Training Precision, Training Recall, Training f1_score, Training areaUnderROC, Training areaUnderPR,
    Testing Precision, Testing Recall, Testing f1_score, Testing areaUnderROC, Testing areaUnderPR
    :param train_predictions: train predicted values.
    :param test_predictions: test actual values.
    :param variable_to_predict: variable_to_predict.
    :param prediction_col: predictionCol
    :return: dictionary with accuracy scores.
    """
    evaluatorMulti = MulticlassClassificationEvaluator(labelCol=variable_to_predict, predictionCol=prediction_col)
    evaluatorRoc = BinaryClassificationEvaluator(
        labelCol=variable_to_predict, rawPredictionCol="prediction", metricName='areaUnderROC')
    evaluatorPR = BinaryClassificationEvaluator(
        labelCol=variable_to_predict, rawPredictionCol="prediction", metricName='areaUnderPR')

    labels = sorted(train_predictions.select(prediction_col).rdd.distinct().map(lambda r: r[0]).collect())
    train_metrics = train_predictions.select(variable_to_predict, prediction_col)
    test_metrics = test_predictions.select(variable_to_predict, prediction_col)

    scores = {}
    for label in labels:
        scores.update({
            'precision (' + str(label) + ') (train set)': evaluatorMulti.evaluate(
                train_metrics, {evaluatorMulti.metricName: "weightedPrecision"}),
            'recall (' + str(label) + ') (test set)': evaluatorMulti.evaluate(
                train_metrics, {evaluatorMulti.metricName: "weightedRecall"}),
            'f1 score (' + str(label) + ') (train set)': evaluatorMulti.evaluate(
                train_metrics, {evaluatorMulti.metricName: "f1"}),
            'accuracy (' + str(label) + ') (train set)': evaluatorMulti.evaluate(
                train_metrics, {evaluatorMulti.metricName: "accuracy"}),
            'areaUnderROC (' + str(label) + ') (train set)': evaluatorRoc.evaluate(train_metrics),
            'areaUnderPR (' + str(label) + ') (train set)': evaluatorPR.evaluate(train_metrics),
            'precision (' + str(label) + ') (test set)': evaluatorMulti.evaluate(
                test_metrics, {evaluatorMulti.metricName: "weightedPrecision"}),
            'recall (' + str(label) + ') (test set)': evaluatorMulti.evaluate(
                test_metrics, {evaluatorMulti.metricName: "weightedRecall"}),
            'f1_score (' + str(label) + ') (test set)': evaluatorMulti.evaluate(
                test_metrics, {evaluatorMulti.metricName: "f1"}),
            'accuracy (' + str(label) + ') (test set)': evaluatorMulti.evaluate(
                test_metrics, {evaluatorMulti.metricName: "accuracy"}),
            'areaUnderROC (' + str(label) + ') (test set)': evaluatorRoc.evaluate(test_metrics),
            'areaUnderPR (' + str(label) + ') (test set)': evaluatorPR.evaluate(test_metrics)
        })
    return scores


def regression(train_predictions, test_predictions, train_actuals=None, test_actuals=None,
               variable_to_predict='target', prediction_col='prediction', *args, **kwargs):
    """
    Return a dictionary of accuracy scores for provided predicted values. The following scores are returned:
    Training Accuracy(R^2), Testing Accuracy(R^2), Training MAPE, testing MAPE.
    :param train_predictions: train predicted values.
    :param test_predictions: test actual values.
    :param variable_to_predict: variable_to_predict.
    :param prediction_col: predictionCol.
    :return: dictionary with accuracy scores.
    """
    reg_eval = RegressionEvaluator(labelCol=variable_to_predict, predictionCol=prediction_col)
    mape_eval = MapeEvaluator(labelCol=variable_to_predict, predictionCol=prediction_col)
    return {
        'R2 (train set)': reg_eval.evaluate(train_predictions, {reg_eval.metricName: 'r2'}),
        'R2 (test set)': reg_eval.evaluate(test_predictions, {reg_eval.metricName: 'r2'}),
        'MAE (train set)': reg_eval.evaluate(train_predictions, {reg_eval.metricName: 'mae'}),
        'MAE (test set)': reg_eval.evaluate(test_predictions, {reg_eval.metricName: 'mae'}),
        'RMSE (train set)': reg_eval.evaluate(train_predictions, {reg_eval.metricName: 'rmse'}),
        'RMSE (test set)': reg_eval.evaluate(test_predictions, {reg_eval.metricName: 'rmse'}),
        'MAPE (train set)': mape_eval.evaluate(train_predictions),
        'MAPE (test set)': mape_eval.evaluate(test_predictions)
    }


class MapeEvaluator(Evaluator):
    """
    A mean absolute percentage error evaluator for spark ML
    """
    def __init__(self, predictionCol="prediction", labelCol="label"):
        super(MapeEvaluator, self).__init__()
        self.predictionCol = predictionCol
        self.labelCol = labelCol

    def _evaluate(self, dataset):
        dataset = dataset.withColumn('non_zero',
                                     F.when(F.col(self.predictionCol) == 0, 1).otherwise(F.col(self.predictionCol)))
        return (dataset.select(F.mean(
                F.abs(
                    (F.col(self.labelCol) - F.col(self.predictionCol)) / F.col('non_zero'))).alias('mape')) \
                .collect()[0][0]) * float(100)

    def isLargerBetter(self):
        return False


class F1ScoreEvaluator(Evaluator):
    """
    An F1 evaluator for spark ML
    """
    def __init__(self, predictionCol="prediction", labelCol="label"):
        super(F1ScoreEvaluator, self).__init__()
        self.predictionCol = predictionCol
        self.labelCol = labelCol

    def _evaluate(self, dataset):
        eval_multi = MulticlassClassificationEvaluator(labelCol=self.labelCol, predictionCol=self.predictionCol)
        return eval_multi.evaluate(dataset.select(self.labelCol, self.predictionCol), {eval_multi.metricName: "f1"})

    def isLargerBetter(self):
        return True




import os
from mlapp.integrations.aml.utils.deploy import get_best_model_in_experiment, deploy_model


def run_script(ws, experiment_name, asset_name, asset_label=None, run_id=None,
               score_metric=None, greater_is_better=False, cpu_cores=1, memory_gb=8):

    if run_id is None:
        # get best model in experiment
        run_id = get_best_model_in_experiment(
            ws, experiment_name, asset_name, asset_label, score_metric, greater_is_better)

    # deployment name
    if asset_label is not None:
        aci_service_name = asset_name.replace('_', '-') + '-' + asset_label.replace('_', '-')
    else:
        aci_service_name = asset_name.replace('_', '-')

    # entry script location
    entry_script = os.path.join("deployment", "aml_deployment.py")

    # deploy
    deploy_model(ws, aci_service_name, experiment_name, asset_name, asset_label, run_id, cpu_cores, memory_gb, entry_script)




import numpy as np
import os
import pandas as pd
from mlapp.utils.features.pandas import evaluate_df_with_binary_output, interact_features
from assets.classification.classification_feature_engineering import ClassificationFeatureEngineering
from mlapp.managers import DataManager, pipeline


class ClassificationDataManager(DataManager):
    # -------------------------------- custom init function ---------------------------------------
    def __init__(self, config, *args, **kwargs):
        DataManager.__init__(self, config, *args, **kwargs)

        # Custom initiates
        if self.data_settings.get('data_handling', None) is None:
            raise Exception("'data_handling' is needed in model configuration file!!!")

        # Feature engineering helper class
        self.feature_engineering_instance = ClassificationFeatureEngineering()

        # actions for continuous features
        self.AUTO_BIN = "auto_bin"
        self.KEEP_AS_IS = "keep_as_is"
        self.REMOVE = "remove"

    # -------------------------------------- train methods -------------------------------------------
    @pipeline
    def load_train_data(self, *args):
        local_path = os.path.join(os.getcwd(), self.data_settings["local_data_csvs"][0].get("path", ""))
        data = self._load_data(local_path)
        return data

    @pipeline
    def clean_train_data(self, data):
        print('------------- CLEAN TRAIN DATA -------------')
        data, missing_values = self._clean_data(data)
        self.save_metadata('missing_values', missing_values)
        return data

    @pipeline
    def transform_train_data(self, data):
        results_df = self._transform_data(data)
        # ------------------------- return final features to train -------------------------
        print("------------------------- return final features to train -------------------------")
        return results_df

    # ------------------------------------- forecast methods -----------------------------------------
    @pipeline
    def load_forecast_data(self,*args):
        print('------------- LOAD FORECAST DATA -------------')
        local_path = os.path.join(os.getcwd(),self.data_settings["local_data_csvs"][0].get("path", ""))
        data = self._load_data(local_path)
        print('------------- DONE LOAD FORECAST DATA -------------')
        return data

    @pipeline
    def clean_forecast_data(self, data):
        print('------------- CLEAN FORECAST DATA -------------')
        data, missing_values = self._clean_data(data)
        return data

    @pipeline
    def transform_forecast_data(self, data):
        data_handling = self.data_settings.get('data_handling', {})
        dates_transformation = data_handling.get("dates_transformation", {})
        categorical_features = self.get_metadata('evaluator_features_mappings')
        return self._transform_data(
            data,
            mandatory_categorical_features=categorical_features,
            extraction_date=dates_transformation.get('extraction_date'))

    @pipeline
    def load_target_data(self, *args):
        raise NotImplementedError()

    def transform_data_for_exploration(self, data):
        data_handling = self.data_settings.get('data_handling', {})

        for col in data.columns:
            if data[col].dtype == np.bool:
                data[col] = data[col].astype(int)

        data_df = self.feature_engineering_instance.transform_features(data, data_handling.get('features_handling'))

        print("------------------------- handling y variable -------------------------")
        final_y = data_df[self.model_settings.get("variable_to_predict", self.data_settings.get("variable_to_predict"))]
        data_df = data_df.drop(self.model_settings.get("variable_to_predict", self.data_settings.get("variable_to_predict")), axis=1)

        print("------------------------- feature interactions -------------------------")
        interactions_list = data_handling.get('features_interactions', [])
        data_df = interact_features(data_df, interactions_list)

        print("------------------------- merge final features with y value -------------------------")
        results_df = data_df.join(final_y.rename(self.model_settings.get("variable_to_predict", self.data_settings.get("variable_to_predict"))))

        # ------------------------- return final features to train -------------------------
        print("------------------------- return final features to train -------------------------")
        return results_df

    # ------------------------------ helper methods for train/forecast --------------------------------
    @staticmethod
    def _load_data(path):
        data = pd.read_csv(path, encoding='ISO-8859-1')
        return data

    def _clean_data(self, data):
        data_handling = self.data_settings.get('data_handling')
        set_features_index = data_handling.get('set_features_index', [])
        if set_features_index:
            data = data.set_index(set_features_index)

        features_for_train = data_handling.get('features_for_train', [])
        features_to_remove = data_handling.get('features_to_remove', [])

        if features_to_remove:
            for feature_to_remove in features_to_remove:
                if feature_to_remove in data.columns and (
                        feature_to_remove != self.model_settings.get("variable_to_predict", self.data_settings.get("variable_to_predict"))):
                    data = data.drop(feature_to_remove, axis=1)

        if features_for_train:
            features_for_train = list(filter(lambda x: x not in features_to_remove, features_for_train))
            features_for_train += [self.model_settings.get("variable_to_predict", self.data_settings.get("variable_to_predict"))]
            data = data[features_for_train]

        # ------------------------- Removing high percentage of null features -------------------------
        print("------------------------- Removing high percentage of null features -------------------------")
        data = self.feature_engineering_instance.remove_features_by_null_threshold(data, data_handling.get(
            'feature_remove_by_null_percentage', 0.3))

        print("------------------------- fill na -------------------------")
        missing_values = self.get_metadata('missing_values', {})
        data, missing_values = self.feature_engineering_instance.fillna_features(
            data, data_handling.get('features_handling', {}), data_handling.get('default_missing_value', 0),
            missing_values)

        return data, missing_values

    def _transform_data(self, data_df, mandatory_categorical_features=None, extraction_date=None):
        data_handling = self.data_settings.get('data_handling', {})
        evaluator_settings = data_handling.get('evaluator_settings', {})

        for col in data_df.columns:
            if data_df[col].dtype == np.bool:
                data_df[col] = data_df[col].astype(int)

        action_for_continuous_features = data_handling.get('action_for_continuous_features', self.AUTO_BIN)

        data_df = self.feature_engineering_instance.transform_features(data_df, data_handling.get('features_handling'))
        dates_transformation = data_handling.get('dates_transformation', {})
        if dates_transformation:
            for col in dates_transformation.get('columns', []):
                data_df = self.feature_engineering_instance.get_days_from_date(
                    data_df, col,
                    dates_transformation.get('extraction_date') if extraction_date is None else extraction_date)

        # ------------------------- handling y variable -------------------------
        print("------------------------- handling y variable -------------------------")
        data_df, final_y = self.feature_engineering_instance.handle_y_variable(
            data_df, self.model_settings.get("variable_to_predict", self.data_settings.get("variable_to_predict")),
            data_handling['y_variable'])

        print("------------------------- feature interactions -------------------------")
        interactions_list = data_handling.get('features_interactions', [])
        data_df = interact_features(data_df, interactions_list)

        # ------------------------- bin some features -------------------------
        print("------------------------- bin continuous features -------------------------")
        data_df = self.feature_engineering_instance.bin_continuous_features(
            data_df, data_handling.get("features_to_bin", []))

        # -------------- Transform and split features to categorical and continues features ----------------------------
        print("---------------------- transform and split features to categorical and continues features -------------")
        data_df, categorical_columns, continuous_columns, binary_columns, continuous_bins = \
            self.feature_engineering_instance.transform_and_split_features_to_categorical_and_continuous(
                data_df,
                data_handling.get("dates_format", ["%d/%m/%Y", "%Y-%m-%d"]),
                action_for_continuous_features == self.AUTO_BIN)

        # TODO: take out the evaluator to a different file
        # ------------------ Call evaluator with continuous features bins and categorical values -----------------------
        print("--------------- Call evaluator with continues features bins and categorical values --------------------")
        store_evaluator_features = evaluator_settings.get("store_evaluator_features", False)

        # non_continuous_columns = binary_columns + categorical_columns + list(continuous_bins.keys())
        non_continuous_columns = categorical_columns + list(continuous_bins.keys())
        evaluator_features_sep = evaluator_settings.get("evaluator_features_separator", '_|_')
        dummies_data_df = pd.get_dummies(data_df, columns=non_continuous_columns, prefix_sep=evaluator_features_sep)

        if mandatory_categorical_features is None and data_handling.get('use_evaluator', True):
            evaluated_df = evaluate_df_with_binary_output(
                data_df[non_continuous_columns], final_y, categorical_features=categorical_columns,
                bins_dict=continuous_bins)

            # store evaluator
            self.save_dataframe('evaluator_features', evaluated_df)

            # -- Evaluator: Filter returned features from evaluator according to filter_evaluator_threshold ------------
            print("---- Evaluator: Filter returned features from evaluator according to filter_evaluator_threshold ---")

            if len(final_y.unique()) == 2:
                y_precision = np.true_divide(final_y.sum(), len(final_y))
            else:
                y_precision = 0

            filter_evaluator_threshold = evaluator_settings.get('filter_evaluator_threshold', .25)
            min_threshold = filter_evaluator_threshold + y_precision
            max_threshold = 1
            diff_threshold = max_threshold - min_threshold
            mid_threshold = (diff_threshold / 2) + min_threshold
            lower_features = evaluated_df[
                (evaluated_df['precision_by_1'] > min_threshold) & (evaluated_df['precision_by_1'] < mid_threshold)]
            higher_features = evaluated_df[
                (evaluated_df['precision_by_1'] >= mid_threshold) & (evaluated_df['precision_by_1'] < max_threshold)]

            # ------------------------- Combine filtered evaluator features -------------------------
            print("------------------------- Combine filtered evaluator features -------------------------")
            combined_lower_features_df, lower_features_mapping = self.feature_engineering_instance.\
                combine_categorical_features(dummies_data_df,lower_features,sep=evaluator_features_sep)

            combined_higher_features_df, higher_features_mapping = self.feature_engineering_instance.\
                combine_categorical_features(dummies_data_df,higher_features,sep=evaluator_features_sep)

            combined_df = pd.concat([combined_lower_features_df, combined_higher_features_df], axis=1)

            # creates evaluator_features_mappings
            evaluator_features_mappings = {**lower_features_mapping,**higher_features_mapping}
            for feature in set(lower_features_mapping.keys()) & set(higher_features_mapping.keys()):
                evaluator_features_mappings[feature].extend(lower_features_mapping[feature])
            self.save_metadata("evaluator_features_mappings", evaluator_features_mappings)
            results_df = combined_df
        elif data_handling.get('use_evaluator', True):
            mandatory_features = pd.DataFrame()
            for key in mandatory_categorical_features:
                for item in mandatory_categorical_features[key]:
                    mandatory_features[item['name']] = \
                        dummies_data_df[key + evaluator_features_sep + item['categories'][0]]
                    for col in item['categories']:
                        mandatory_features[item['name']] = \
                            mandatory_features[item['name']] | dummies_data_df[key + evaluator_features_sep + col]

            results_df = mandatory_features
        else:
            results_df = data_df

        if action_for_continuous_features == self.AUTO_BIN or action_for_continuous_features == self.REMOVE:
            pass
        elif action_for_continuous_features == self.KEEP_AS_IS:
            data_df.drop(non_continuous_columns, axis=1, inplace=True)
            if isinstance(results_df, pd.DataFrame):
                results_df = pd.concat([data_df, results_df], axis=1)
            else:
                results_df = data_df

        # ------------------------- merge final features with y value -------------------------
        print("------------------------- merge final features with y value -------------------------")
        results_df = results_df.join(final_y.rename(
            self.model_settings.get("variable_to_predict", self.data_settings.get('variable_to_predict'))))
        return results_df




from mlapp.utils.exceptions.base_exceptions import FrameworkException


class SkipServiceException(FrameworkException):
    pass


class SkipToLocalException(FrameworkException):
    pass


class MissingConnectionException(FrameworkException):
    pass


class UnsupportedFileType(FrameworkException):
    pass


class DataFrameNotFound(FrameworkException):
    pass


class AutoMLException(FrameworkException):
    pass



import os
from azureml.core import Environment
from azureml.core.environment import DEFAULT_CPU_IMAGE


def get_mlapp_environment(workspace, env_name, version=None):
    return Environment.get(workspace=workspace, name=env_name, version=version)


def create_env_from_requirements(file_path='requirements.txt', name='mlapp', endpoint=False):
    env = Environment.from_pip_requirements(name=name, file_path=file_path)

    # Enable Docker
    env.docker.enabled = True

    # Set Docker base image to the default CPU-based image
    path_to_dockerfile = os.path.join(os.getcwd(), 'deployment', 'Dockerfile')
    if os.path.exists(path_to_dockerfile):
        # Set custom Docker image
        env.docker.base_image = None
        env.docker.base_dockerfile = path_to_dockerfile
    else:
        # Set Docker base image to the default CPU-based image
        env.docker.base_image = DEFAULT_CPU_IMAGE

    # Use conda_dependencies.yml to create a conda environment in the Docker image for execution
    env.python.user_managed_dependencies = False

    if endpoint:
        env.inferencing_stack_version = 'latest'

    return env


def display_mlapp_environments(workspace, name='mlapp'):
    envs = Environment.list(workspace=workspace)

    for env in envs:
        if env.startswith(name):
            print("Name", env)
            print("packages", envs[env].python.conda_dependencies.serialize_to_string())


from mlapp.managers.user_managers import ModelManager
from mlapp.managers.user_managers import DataManager
from mlapp.managers.pipeline_manager import pipeline

__all__ = ('ModelManager',
           'DataManager',
           'pipeline')



from mlapp.mlapp_cli.common.cli_utilities import to_lower, is_int, to_int, to_upper, is_positive, clean_spaces

steps = {
    'number_of_steps': {
        "display_name": "Choose how many pipeline steps to deploy",
        "short_description": 'Press enter to set number of steps to 2',
        "validations": [is_int, is_positive],
        "transformations": [to_int],
        "error_msg": "Number of steps should b a valid positive integer.",
        "default": 2,
        "required": True,
        "nested_dependency": {
            "name": {
                "display_name": "compute target",
                "short_description": 'Press enter compute-target name',
                "transformations": [to_lower],
                "error_msg": "compute_target should be exists in your Workspace.",
                "required": True
            },
            "type": {
                "display_name": "Is it databricks machine?",
                "short_description": 'Y/N, default No',
                "transformations": [to_lower, clean_spaces],
                "values": {
                    'y': 'databricks',
                    'n': 'aml_compute_target',
                    'yes': 'databricks',
                    'no': 'aml_compute_target'
                },
                "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
                "default": 'aml_compute_target',
                "required": True
            },
            "set_advanced": {
                "display_name": "Do you want to set compute-target advanced settings?",
                "short_description": 'Y/N, default no will use compute-target default settings',
                "transformations": [to_lower, clean_spaces],
                "values": {
                    'y': True,
                    'n': False,
                    'yes': True,
                    'no': False
                },
                "error_msg": "Possible values should be 'y', 'n', 'yes' or 'no'.",
                "default": 'false',
                "required": True
            },
            "vm_size": {
                "display_name": "vm size",
                "short_description": 'Press enter to set vm size to \'STANDARD_D2_V2\'',
                "transformations": [to_upper],
                "error_msg": "Port should contain digits only.",
                "default": 'STANDARD_D2_V2',
                "required": False
            },
            "min_nodes": {
                "display_name": "min nodes",
                "short_description": 'Press enter to set minimum nodes to 0',
                "validations": [is_int, is_positive],
                "transformations": [to_int],
                "error_msg": "Min nodes should be a valid non negative integer.",
                "default": 0,
                "required": False
            },
            "max_nodes": {
                "display_name": "max nodes",
                "short_description": 'Press enter to set maximum nodes to 4',
                "validations": [is_int, is_positive],
                "transformations": [to_int],
                "error_msg": "Max nodes should be a valid non negative integer.",
                "default": 4,
                "required": False
            },
            "idle_seconds_before_scale_down": {
                "display_name": "idle seconds before scale down",
                "short_description": 'Press enter to idle seconds before scale down to 120',
                "validations": [is_int, is_positive],
                "transformations": [to_int],
                "error_msg": "idle seconds before scale down should be a valid non negative integer.",
                "default": 120,
                "required": False
            }
        }
    }
}



basic_regression_config = {
    "pipelines_configs": [{
        "data_settings": {
            "local_file_path": "data/diabetes.csv",  # data source
            "variable_to_predict": "target",  # column name of variable to predict
            # features handling configurations
            "data_handling": {
                "features_for_train": [],  # leave empty for selecting all features
                "set_features_index": [],  # leave empty for indexing by row index
                # features to remove
                "features_to_remove": ["sex"],
                "feature_remove_by_null_percentage": 0.3
            }
        },
        # task settings
        "job_settings": {
            "asset_name": "basic_regression",
            "pipeline": ['load_train_data', 'clean_train_data', 'cache_features']
        }
    }]
}



import numpy as np

# train configurations
spark_classification_config = {
    "pipelines_configs": [{
        "data_settings": {
            "model_name": "breast_cancer",
            "local_data_csvs": [{
                "name": "breast_cancer",
                "path": "data/breast_cancer.csv"
            }],
            "variable_to_predict": "answer",  # column name of variable to predict
            # features handling configurations
            "data_handling": {
                # y variable feature engineering
                "y_variable": {
                    "type": "binary",  # binary/multi/continuous
                    "categories_labels": ["NEGATIVE", "POSITIVE"],  # category labels
                    "continuous_to_category_bins": [-1, 1, 2],
                    # bins values in case there is a need to cut the y to categories
                    "label_to_predict": ["POSITIVE"]  # target label to predict
                },
                "features_for_filter": {},  # leave empty for not filter
                "features_for_train": None,  # leave empty for selecting all features
                "set_features_index": None,  # leave empty for indexing by row index
                "features_to_remove": ["texture error", "area error", "smoothness error", "compactness error",
                                       "concave points error", "symmetry error", "worst smoothness",
                                       "worst compactness", "worst concavity", "worst concave points", "worst symmetry",
                                       "worst fractal dimension"],
                "feature_remove_by_null_percentage": 0.3,
                "dates_format": "yyyyMMdd",
                "default_missing_value": 0,
                "features_handling": {
                    "mean radius": {"fillna": "mean", "transformation": ["lambda a: F.pow(a, 2)", "F.sqrt"]},
                    "radius error": {"fillna": 0, "transformation": []}
                },
                "dates_transformation": {"extraction_date": "20180430", "columns": []},
                "features_to_bin": [
                    {"name": "mean radius", "bins": [12.3, 15.3]},
                    {"name": "mean texture", "bins": [15, 23]},
                    {"name": "mean perimeter", "bins": [72, 109]},
                    {"name": "mean area", "bins": [361, 886]},
                    {"name": "mean smoothness", "bins": [0.074, 0.11]},
                    {"name": "mean compactness", "bins": [0.047, 0.137, 0.228]},
                    {"name": "mean concavity", "bins": [0.023, 0.12]},
                    {"name": "mean concave points", "bins": [4]},
                    {"name": "mean symmetry", "bins": ['quantile']},
                    {"name": "mean fractal dimension", "bins": [0.0518, 0.0541, 0.0742, 0.0827]},
                    {"name": "radius error", "bins": [0.19, 0.56, 0.83]},
                    {"name": "worst area", "bins": [500, 1050]},
                    {"name": "worst perimeter", "bins": [85, 120]},
                    {"name": "worst texture", "bins": [16.6, 42]},
                    {"name": "worst radius", "bins": [12.5, 18]},
                    {"name": "fractal dimension error", "bins": [0.006, 0.0135]},
                    {"name": "concavity error", "bins": [0.011, 0.082, 0.15]},
                    {"name": "perimeter error", "bins": [1.3, 5]}
                ],
                "features_to_convert_to_dummies": ["mean radius", "mean texture", "mean perimeter", "mean area",
                                                   "mean smoothness", "mean compactness", "mean concavity",
                                                   "mean concave points", "mean symmetry", "mean fractal dimension",
                                                   "radius error", "worst area", "worst perimeter", "worst texture",
                                                   "worst radius", "fractal dimension error", "concavity error",
                                                   "perimeter error"],
                "bin_continuous_features": True,
                "auto_get_dummies": True,
                "max_categories_num": 10,
            },
            "data_output_keys": {"results": ["missing_values", "evaluator_features_mappings"], "images_files": []},
        },
        "model_settings": {
            "variable_to_predict": "answer",
            "label_to_predict": ["POSITIVE"],  # target label to predict
            "scale": False,
            "train_percent": 0.8,
            "auto_ml":{
                "binary": {
                    "models": ["LogisticRegression", "SVC", "RandomForestClassifier"],
                    "hyper_params": {
                        'LogisticRegression': {
                            'elasticNetParam': list(np.linspace(0, 1, 2)),
                            'regParam': list(np.linspace(0.01, 1, 2))
                        },
                        'SVC': {
                            'regParam': list(np.linspace(0.01, 1, 2))
                        },
                        'RandomForestClassifier': {
                            'maxDepth': [5, 3],
                            'numTrees': [10, 20],
                            'featureSubsetStrategy': ['auto', 'all', 'log2']
                        }
                    }
                },
                "feature_selection": [
                    {
                        "method": "AllFeatures"
                    },
                    {
                        "method": "ChiSqSelector",
                        "params": {
                            "labelCol": "answer",
                            "numTopFeatures": 10
                        }
                    }
                ],
            }
        },
        "job_settings": {
            "asset_name": "spark_classification",
            "pipeline": "train",
            "store_features": False,
        },
    }]
}



import functools
import time
import pika.exceptions
import pika
import ssl
from mlapp.handlers.message_queues.message_queue_interface import MessageQueueInterface


class RabbitMQHandler(MessageQueueInterface):
    def __init__(self, settings):
        """
        Initializes the RabbitMQHandler with it's special connection string
        :param settings: settings from `mlapp > config.py` depending on handler type name.
        """
        super(RabbitMQHandler, self).__init__()

        connection_params = {
            'host': settings.get('hostname'),
            'port': settings.get('port')
        }

        if settings.get('cert_path', False):
            context = ssl.create_default_context(
                cafile=settings['cert_path']
            )
            ssl_options = pika.SSLOptions(context, settings.get('hostname'))
            credentials = pika.PlainCredentials(settings.get('username'), settings.get('password'))
            self.params = pika.ConnectionParameters(
                **connection_params, credentials=credentials, ssl_options=ssl_options)
        else:
            self.params = pika.ConnectionParameters(**connection_params)

        self.connection_timeout = settings.get('connection_timeout', 15)

    def send_message(self, queue_name, body):
        """
        Sends message to the queue
        :param queue_name: name of the topic/queue to send the message to
        :param body: message as string or bytes
        """
        while True:
            connection = None
            channel = None
            try:
                # opening connection
                connection = pika.BlockingConnection(parameters=self.params)
                channel = connection.channel()

                # connect to queue
                channel.queue_declare(queue=queue_name, durable=True)
                channel.confirm_delivery()

                # sending message
                channel.basic_publish(exchange='', routing_key=queue_name, body=body)
                break
            except pika.exceptions.AMQPChannelError as err:
                print("Caught a channel error: {}, stopping...".format(err))
                break
                # recover on all other connection errors
            except pika.exceptions.AMQPConnectionError:
                print("Connection was closed, retrying...")
                time.sleep(1)
                continue
            except Exception as e:
                print(e)
                time.sleep(1)
                continue
            finally:
                # closing connection
                if connection and connection.is_open and channel:
                    channel.close()
                    connection.close()

    def listen_to_queues(self, queue_names, callback):
        """
        Listen to queues/topics
        :param queue_names: list of queue/topic names to listen to
        :param callback: function to call upon receiving a message
        """
        while True:
            try:
                # connection to rabbitMQ
                conn = pika.BlockingConnection(parameters=self.params)
                chan = conn.channel()

                # preparing listen to queues
                on_message_callback = functools.partial(self._on_message, args=(conn, callback))

                for queue in queue_names:
                    chan.queue_declare(queue=str(queue), durable=True)
                    chan.basic_qos(prefetch_count=1)
                    chan.basic_consume(str(queue), on_message_callback)
                    print('[*] Waiting for messages in ' + str(queue) + '. To exit press CTRL+C')

                # listening to queues
                try:
                    chan.start_consuming()
                except KeyboardInterrupt:
                    chan.stop_consuming()
                    conn.close()
                    break

            except pika.exceptions.AMQPChannelError as err:
                print("Caught a channel error: {}, stopping...".format(err))
                break
                # recover on all other connection errors
            except pika.exceptions.AMQPConnectionError:
                print("Connection was closed, retrying...")
                time.sleep(1)
                continue
            except Exception as e:
                print(e)
                time.sleep(1)
                continue

    @staticmethod
    def _ack_message(body, channel, delivery_tag):
        """Note that `channel` must be the same pika channel instance via which
        the message being ACKed was retrieved (AMQP protocol constraint).
        """
        if channel.is_open:
            channel.basic_ack(delivery_tag)
        else:
            # Channel is already closed, so we can't ACK this message;
            # log and/or do something that makes sense for your app in this case.
            print("Channel was closed during the process of this task. Message can't be acknowledged.")

    @staticmethod
    def _on_message(channel, method_frame, header_frame, body, args):
        (connection, callback) = args
        # auto ack is set to true so no need to ack message
        # self._ack_message(body, channel, method_frame.delivery_tag)

        if channel.is_open:
            channel.basic_ack(method_frame.delivery_tag)
        else:
            # Channel is already closed, so we can't ACK this message;
            # log and/or do something that makes sense for your app in this case.
            print("Channel was closed during the process of this task. Message can't be acknowledged.")

        connection.close()
        callback(body)



MAX_STRING_LENGTH = 31
OUTPUTS_FOLDER = "outputs"
AML_MLAPP_FOLDER = "mlapp"
MAX_AML_PIPELINE_STEPS = 10
CONFIG_ARGUMENT = 'config'
INPUT_DATA_ARGUMENT = 'input'
OUTPUT_DATA_ARGUMENT = 'output'
ARG_TYPES = [
    {'name': CONFIG_ARGUMENT, 'output_key': 'config_str', 'help': 'configuration', 'type': 'str'},
    {'name': INPUT_DATA_ARGUMENT, 'output_key': 'input_dirs', 'help': 'input directory', 'type': 'list'},
    {'name': OUTPUT_DATA_ARGUMENT, 'output_key': 'output_dirs', 'help': 'output directory', 'type': 'list'}
]
OUTPUT_PATH_ON_COMPUTE = 'results'
DATA_REFERENCE_NAME = 'results_dir'
PARSED_ARG_CONFIG = 'config_str'
PARSED_ARG_INPUT_DIR = 'input_dirs'
PARSED_ARG_OUTPUT_DIR = 'output_dirs'



basic_regression_config = {
    "pipelines_configs": [
        {
            # train config
            "data_settings": {
                "local_file_path": "data/diabetes.csv",  # data source
                "variable_to_predict": "target",  # column name of variable to predict
                # features handling configurations
                "data_handling": {
                    "features_for_train": [],  # leave empty for selecting all features
                    "set_features_index": [],  # leave empty for indexing by row index
                    # features to remove
                    "features_to_remove": ["sex"],
                    "feature_remove_by_null_percentage": 0.3
                }
            },
            # model settings
            "model_settings": {},
            # task settings
            "job_settings": {
                "asset_name": "basic_regression",
                "pipeline": "feature_engineering"
            }
        }
    ]
}



import inspect
import time
from abc import ABC, abstractmethod
from copy import deepcopy
from mlapp.utils.exceptions.framework_exceptions import AutoMLException


class _AutoMLBase(ABC):
    @abstractmethod
    def _validate_input(self, estimator_family, train, test, *args, **kwargs):
        pass

    @abstractmethod
    def _default_fixed_params(self, estimator_family):
        pass

    @abstractmethod
    def _default_hyper_params(self, estimator_family):
        pass

    @abstractmethod
    def _default_model_classes(self, estimator_family):
        pass

    def _get_models_names(self, estimator_family, models=None):
        if models:
            return models
        else:
            return list(self._default_model_classes(estimator_family).keys())

    @abstractmethod
    def _eval_config_input(self, config_input):
        pass

    def construct_dictionary(self, input_dict, default_dict):
        copy_dict = deepcopy(default_dict)
        if input_dict and isinstance(input_dict, dict):
            copy_dict.update(self._eval_config_input(input_dict))

        return copy_dict

    @abstractmethod
    def _get_model_intercept_and_coefficients(self, *args, **kwargs):
        pass

    @abstractmethod
    def _get_cv_results(self, estimator_family, searches, features, scoring, greater_is_better, *args, **kwargs):
        pass

    @abstractmethod
    def _inner_search(self, estimator_family, train, test, model_key, model_class, fixed_params, hyper_params, cv,
                      scoring, greater_is_better, *args, **kwargs):
        pass

    @abstractmethod
    def _visualization_methods(self, method):
        pass

    def _create_images(self, methods, *args, **kwargs):
        figs = {}
        for method in methods:
            vis_f = self._visualization_methods(method)
            if not vis_f:
                print("INFO: Visualization method '{0}' doesn't exist ".format(method))
                continue
            signature = inspect.signature(vis_f)
            f_args = list(signature.parameters)
            parameters = {k: v.default for k, v in signature.parameters.items() if
                          v.default is not inspect.Parameter.empty}
            passed_parameters = {arg: kwargs[arg] for arg in f_args if arg in kwargs}
            parameters.update(passed_parameters)

            missing_params = set(f_args) - set(parameters.keys())
            if len(missing_params) > 0:
                print(
                    "INFO: Failed to create visualization {0}. Missing Params: {1}".format(method, str(missing_params)))
            else:
                figure = vis_f(*[parameters[arg] for arg in f_args])
                if figure is not None:
                    figs[method] = figure
        return figs

    @abstractmethod
    def _full_data_model_fit(self, model, train, test, *args, **kwargs):
        pass

    @abstractmethod
    def _predict_train_test(self, model, train, test, *args, **kwargs):
        pass

    @abstractmethod
    def _get_scorer(self, estimator_family, metric=None, *args, **kwargs):
        pass

    def _construct_run_params(self, estimator_family, fixed_params, hyper_params, model_classes):
        return self.construct_dictionary(fixed_params, self._default_fixed_params(estimator_family)), \
               self.construct_dictionary(hyper_params, self._default_hyper_params(estimator_family)), \
               self.construct_dictionary(model_classes, self._default_model_classes(estimator_family))

    def _load_feature_selection_method(self, method):
        if callable(method):
            return method
        elif isinstance(method, str) and method != 'AllFeatures':
            globals_method = self._feature_selection_globals(method)
            if not globals_method:
                print(f"INFO: feature selection method '{method}' not in scope.")
            return globals_method

    @abstractmethod
    def _feature_selection_globals(self, method):
        pass

    @abstractmethod
    def _run_feature_selection(self, feature_selection, train, test, *args, **kwargs):
        pass

    @staticmethod
    def _get_feature_selection_methods(feature_selection):
        return feature_selection if feature_selection else [{"method": "AllFeatures"}]

    @staticmethod
    def _get_feature_selection_name(method):
        if isinstance(method, str):
            return method
        else:
            return method.__class__.__name__

    @abstractmethod
    def _scores_function(self, scores_summary_function, estimator_family):
        pass

    def run(self, results, scores_summary, feature_selections, estimator_family, train, test, models=None,
            fixed_params=None, hyper_params=None, model_classes=None, scoring=None, greater_is_better=True, *args,
            **kwargs):

        # validate input
        self._validate_input(estimator_family, train, test, *args, **kwargs)
        # init params
        fixed_params, hyper_params, model_classes = \
            self._construct_run_params(estimator_family, fixed_params, hyper_params, model_classes)

        for feature_selection_method in self._get_feature_selection_methods(feature_selections):
            # feature selection
            selected_features, train_filtered, test_filtered = \
                self._run_feature_selection(feature_selection_method, train, test, *args, **kwargs)

            # run hyper param search per model
            hyper_params_searches = {}
            for model_key in self._get_models_names(estimator_family, models):
                try:
                    model_class = model_classes.get(model_key)
                    if not model_class:
                        print(f"WARNING: `{model_key}` is not supported by MLApp's AutoML. "
                              f"Check naming is correct and that there is no missing python library installation.")
                        continue

                    start_time = time.time()

                    hyper_params_searches[model_key] = self._inner_search(
                        estimator_family, train_filtered, test_filtered, model_key, model_class, fixed_params,
                        hyper_params, scoring, greater_is_better, *args, **kwargs)

                    print("INFO: hyper params search took %.2f seconds for model: %s parameter settings." %
                          ((time.time() - start_time), model_key))

                except Exception as err:
                    raise AutoMLException(
                        f"GridSearch/RandomSearch failed with '{model_key}' model with error: {str(err)}.")

            if len(hyper_params_searches.keys()) == 0:
                raise AutoMLException(f"AutoML has no output! "
                                      f"Check if you are running at least one model or whether all models failed.")

            # get cv results
            cv_results, best_model_key, best_model, best_cv_score, intercept, coefficients = \
                self._get_cv_results(
                    estimator_family, hyper_params_searches, selected_features, scoring, greater_is_better, *args,
                    **kwargs)

            # add figures
            output_plots = {}
            if kwargs.get('visualizations'):
                output_plots = self._create_images(
                    kwargs['visualizations'], hyper_params=hyper_params.get(best_model_key, {}),
                    grid_search_results=hyper_params_searches[best_model_key], greater_is_better=greater_is_better,
                    X_train=train_filtered, X_test=test_filtered, model=best_model, *args, **kwargs)

            # get train/test y_hat
            train_y_hat, test_y_hat = self._predict_train_test(
                best_model, train_filtered, test_filtered, *args, **kwargs)

            # fit model on full train+test data
            if kwargs.get('best_model_full_fit'):
                best_model = self._full_data_model_fit(best_model, train, test, *args, **kwargs)

            if best_model is None:
                raise AutoMLException("Error: AutoML run must pass 'models' with at least one model!")

            # append results
            scores_function = self._scores_function(scores_summary, estimator_family)
            results.add_cv_run(scores_function, estimator_family,
                               self._get_feature_selection_name(feature_selection_method.get('method')), train_y_hat,
                               test_y_hat, selected_features, best_model, intercept, coefficients, best_cv_score,
                               cv_results, output_plots, *args, **kwargs)

        # return results
        return results



import argparse
import os
from mlapp.integrations.aml.utils.constants import MAX_AML_PIPELINE_STEPS, ARG_TYPES, \
    PARSED_ARG_CONFIG, PARSED_ARG_INPUT_DIR, PARSED_ARG_OUTPUT_DIR
from mlapp.utils.general import load_pickle_to_object, save_object_to_pickle


def parse_args():
    parser = argparse.ArgumentParser()
    parsed_args = {
        PARSED_ARG_CONFIG: '{}',
        PARSED_ARG_INPUT_DIR: [],
        PARSED_ARG_OUTPUT_DIR: []
    }

    # add arguments for parse
    for i in range(MAX_AML_PIPELINE_STEPS):
        for arg_type in ARG_TYPES:
            parser.add_argument(
                ('--' + arg_type['name'] + str(i)),
                type=str, dest=(arg_type['name'] + str(i)), help=arg_type['help'])

    # parse arguments
    args = parser.parse_args()

    # save parsed arguments
    for i in range(MAX_AML_PIPELINE_STEPS):
        for arg_type in ARG_TYPES:
            if getattr(args, arg_type['name'] + str(i)) is not None:
                if arg_type['type'] == 'str':
                    parsed_args[arg_type['output_key']] = getattr(args, arg_type['name'] + str(i))
                elif arg_type['type'] == 'list':
                    parsed_args[arg_type['output_key']].append(getattr(args, arg_type['name'] + str(i)))

    return parsed_args


def get_job_file_name(directory):
    return os.path.join(directory, 'output.pkl')


def load_job_output(outputs, directory):
    if os.path.exists(os.path.join(os.getcwd(), get_job_file_name(directory))):
        outputs['input_from_predecessor'] += \
            load_pickle_to_object(os.path.join(os.getcwd(), get_job_file_name(directory)))
    else:
        outputs['input_from_predecessor'] += [None]


def is_flow_summary(config):
    return config.get('flow_config', {}) != {}


def flow_setup(current_run_id, config, input_dirs):
    jobs_outputs = {
        'flow_id': current_run_id,
        'input_from_predecessor': [None]
    }
    is_flow_summarizer = is_flow_summary(config)

    if is_flow_summarizer:
        # load all jobs outputs
        for input_dir in input_dirs:
            load_job_output(jobs_outputs, input_dir)
    else:
        # load last job
        if len(input_dirs) > 0:
            jobs_outputs['input_from_predecessor'] = []
            load_job_output(jobs_outputs, input_dirs[-1])

    return jobs_outputs


def flow_postprocess(config, output_data, output_dirs):
    if not is_flow_summary(config):
        for output_dir in output_dirs:
            file_name = get_job_file_name(output_dir)
            file_path = os.path.join(os.getcwd(), file_name)
            os.makedirs(os.path.join(os.getcwd(), output_dir), exist_ok=True)
            save_object_to_pickle(output_data, os.path.join(os.getcwd(), file_path))
